;; Generated package description from evil-args.el  -*- no-byte-compile: t -*-
(define-package "evil-args" "1.1.0.20240209.210417" "Motions and text objects for delimited arguments in Evil." '((evil "1.0.8")) :commit "a8151556f63c9d45d0c44c8a7ef9e5a542f3cdc7" :authors '(("Connor Smith" . "wconnorsmith@gmail.com")) :maintainer '("Connor Smith" . "wconnorsmith@gmail.com") :keywords '("evil" "vim-emulation") :url "http://github.com/wcsmith/evil-args")
